module.exports = function(grunt) {
  var tarima = require('tarima'),
      _ = grunt.util._;

  grunt.registerMultiTask('tarima', 'Compile Tarima templates into JavaScript or plain output', function() {

    var options = this.options({
      callback: null,
      join: false,
      includes: {},
      locals: {},
      cwd: ''
    });

    _.each(options.includes, function(all, type) {
      _.each(all, function(file, index) {
        options.includes[type][index] = grunt.file.read(file);
      });
    });

    this.files.forEach(function (f) {
      var validFiles = removeInvalidFiles(f);

      if ('function' === typeof options.filter) {
        validFiles = options.filter(validFiles);
      }

      writeFile(f.dest, concatOutput(validFiles, options), options);
    });
  });

  var removeInvalidFiles = function(files) {
    return files.src.filter(function(filepath) {
      if (!grunt.file.exists(filepath)) {
        grunt.verbose.warn('Source file "' + filepath + '" not found.');
        return false;
      } else {
        var exts = filepath.split('/').pop().split('.');

        if (exts.length < 2) {
          return false;
        }

        if (['jade', 'ract', 'less', 'hbs', 'us'].indexOf(exts[1]) > -1) {
          grunt.verbose.writeln('Source file "' + filepath + '" skipped.');
          return false;
        }

        return true;
      }
    });
  };

  var concatOutput = function(files, options) {
    return files.map(function(filepath) {
      var peak = filepath.split('.').pop(),
          params = _.omit(options, 'filter', 'callback', 'join');

      if (peak === filepath) {
        return;
      }

      if (filepath.indexOf(':') > -1) {
        var code = grunt.file.read(filepath.split(':')[0]);
        return compileTarima(code, params, filepath.split(':')[1]);
      }

      var code = grunt.file.read(filepath);
      return compileTarima(code, params, filepath);
    });
  };

  var compileTarima = function(code, options, filepath) {
    try {
      return tarima.parse(filepath, code, options);
    } catch (e) {
      grunt.fail.warn('Tarima failed to compile.');
    }
  };

  var warnOnEmptyFile = function (path) {
    grunt.log.warn('Destination "' + path + '" not written because compiled files were empty.');
  };

  var writeFile = function (path, params, options) {
    var filepath, partial, code, next;
    var views = {}, exts = [], plain;

    for (partial in params) {
      partial = params[partial];

      if ('function' === typeof options.process) {
        partial = options.process(partial);
      }

      if ('object' !== typeof partial) {
        continue;
      }

      next = partial.params.parts[0];
      plain = 'js' !== partial.params.ext;
      code = partial.compile(options.locals);

      if (!code || code.length < 1) {
        warnOnEmptyFile(path);
      } else {
        code = code.replace(/\n{2,}/g, '\n');
        next = next || partial.params.ext;

        if (options.join) {
          if (exts.indexOf(next) === -1) {
            exts.push(next);
          }

          views[partial.params.keypath] = code;
        } else {
          path = path.substr(-1) === '/' ? path : path + '/';

          if (options.cwd) {
            filepath = partial.params.filepath.substr(options.cwd.length).replace(/^\//, '');
            filepath = filepath ? filepath + '/' + partial.params.name : partial.params.name;
            filepath = path + filepath;
          } else {
            filepath = path + partial.params.name;
          }

          if (plain) {
            filepath += '.' + partial.params.ext;
            grunt.file.write(filepath, code);
          } else {
            filepath += '.' + partial.params.ext;
            grunt.file.write(filepath, getTpl(code, getDeps(next), options.callback) + '\n');
          }

          grunt.log.ok('Source file "' + filepath + '" created.');
        }
      }
    }

    if (options.join) {
      if (plain) {
        code = _.values(views).join('\n');
      } else {
        code = getTpl(views, _.map(_.unique(exts), function(value) { return getDeps(value); }).join(''), options.callback);
      }

      grunt.file.write(path, code);
      grunt.log.ok('File ' + path + ' created.');
    }
  };

  var fnWrapper = function(val) {
    return /^\s*(function[\s(]|Handlebars\.template)/.test(val) ? val : 'function(){' + val.replace(/^\s+|\s+$/g, '') + '}';
  };

  var fixDep = function(type) {
    var map = {
      hbs: 'handlebars-runtime',
      jade: 'jade-lib-runtime'
    };

    return map[type] || type;
  };

  var getDeps = function (type) {
    var prefix = '',
        context = "('undefined' === typeof window ? global : window)";

    var html_escape = function(s) {
          return s !== null ? String(s).replace(/[&<>"']/g, function(m) {
            return ({ '&': '&amp;', '<': '&lt;', '>': '&gt;', '"': '&quot;', "'": '&#39;' })[m];
          }) : '';
        };

    switch (type) {
      case 'less':
      case 'jade':
        prefix = "var " + type + " = " + context + "." + type + " || require('" + fixDep(type) + "');";
      break;
      case 'hbs':
        prefix = "var Handlebars = " + context + ".Handlebars || require('" + fixDep(type) + "');";
      break;
      case 'us':
        prefix = "var _ = " + context + "._ || { escape: " + html_escape.toString() + " };";
      break;
    }

    return prefix;
  };

  var getTpl = function(tpl, deps, callback) {
    var output = [],
        views = [];

    if ('function' === typeof callback) {
      return callback(tpl, deps);
    }

    output.push(deps);
    output.push('\nmodule.exports = ');

    if ('object' === typeof tpl) {
      output.push('{');

      _.each(tpl, function(value, key) {
        views.push("\n'" + key + "': " + fnWrapper(value));
      });

      output.push(views.join(','));
      output.push('}');
    } else {
      output.push(fnWrapper(tpl));
    }

    return output.join('') + ';';
  };
};
